package util;

import com.apple.itunes.*;
import com.jacob.activeX.ActiveXComponent;
import com.jacob.com.ComThread;
import com.jacob.com.Dispatch;
import com.jacob.com.Variant;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

public class ITunesManager {

    private static Log LOG = LogFactory.getLog(ITunesManager.class);

    public static void addTrackToItunes(String fileLocation, String stationName) {
        LOG.info("attempting to communicate with itunes - please make sure it is running");
        System.setProperty("java.library.path", "bin");
        ComThread.InitMTA(true);
        ActiveXComponent iTunesCom = new ActiveXComponent("iTunes.Application");
        Dispatch iTunesController = iTunesCom.getObject();
        IiTunes it = new IiTunes(iTunesController);
        IITLibraryPlaylist libraryPlaylist = it.getLibraryPlaylist();
        IITOperationStatus iitOperationStatus = libraryPlaylist.addFile(fileLocation);

        while (iitOperationStatus.getInProgress()) {
            System.out.println("waiting for track to be added");
        }

        IITTrackCollection tracks = iitOperationStatus.getTracks();
        IITTrack item = tracks.getItem(1);
        LOG.info("track: " + item.getName() + " added to itunes libray");

        if (stationName != null) {
            LOG.info("adding track to playlist: " + stationName);
            IITPlaylist playListByName = findPlayListByName(stationName, it, true);
            if (playListByName != null) {
                playListByName.addTrack(new Variant(item));
                IITTrackCollection pt = playListByName.getTracks();
                IITTrack itemByName = pt.getItemByName(item.getName());
                itemByName.play();
                LOG.info("track: " + item.getName() + " added successfully to playlist: " + stationName);
            } else {
                LOG.info("unable to add track: " + item.getName() + " to playlist: " + stationName);
            }
        }
        //item.play();
        ComThread.Release();
    }

    private static IITPlaylist findPlayListByName(String stationName, IiTunes itunes, boolean create) {
        IITSourceCollection sourceList = itunes.getSources();
        for (int i = 1; i <= sourceList.getCount(); i++) {
            IITSource s = sourceList.getItem(i);
            System.out.println(s.getClass().getName() + "|" + s.getName());
            IITPlaylistCollection lists = s.getPlaylists();
            for (int j = 1; j <= lists.getCount(); j++) {
                IITPlaylist pl = lists.getItem(j);
                if (pl.getName().equals(stationName)) {
                    LOG.info("station playlist: " + stationName + " already exist");
                    return pl;
                }
            }
            if (create) {
                LOG.info("creating station playlist: " + stationName);
                if (s.getName().equals("Library")) {
                    return itunes.createPlaylist(stationName);
                }
            }
            else {
                LOG.debug("playlist: " + stationName + " not found.");
            }

        }
        return null;
    }
}
